#include <stdio.h>
#include <assert.h>
#include "sk.h"
#include "verr.h"
/*
 * skatule -- shedule robots between moving robots
 *
 * au: vitas@popelka.ms.mff.cuni.cz
 * cp: gpl
 */

int t_vac::id_top = 0;
int t_needle::id_top = 0;
double t_world::disk = .1;
double t_world::disk_sq = t_world::disk * t_world::disk;

/*
 * dijkstra: 
 *
 * v kazde `neele' je seznam `vacancy' prazdnych mist kdy se tam
 * muze. startovni vacancy, je pridana do `vac_bag'u (priorityQ dle
 * nejmensiho casu).  
 * 
 * nejmensi vacancy je vybrana, jsou prozkoumavany vsechny "sousedske"
 * needle.
 *
 * najde se min cas ve ktery mohu byt nasmerovan smerem na tuto neele.
 * (`src_to_dest_angle')
 *
 * najde se minmalni cas ve ktery mohu byt v kazde vacanci teto
 * needle.  (*)
 * 
 * vsechny vakance ktere byli vylepseny, jsouu pridany do `vac_bag'u
 *
 * protoze neplati trouhelnikova nerovnost musim skoncit az je
 * nejnizsi polozka `vac_bag'u  mensi nez nejnizsi dosazeny cas u cile
 *
 * (*) spocitam teoreticky interval kdy se mohu na danou needle dostat
 * tj. <src_to_dest_angle,end>+dist/rychlost.
 * 
 * pro kazdou vacanci na cilove needli v tomto intervalu:
 * spocitam teotetickou casovou hranici v pocatecni needli. zkousim
 * s krokem .1sec trasu  start cil. prvni ktera uspeje, zjemnim
 * s krokem .05
 *
 * update vacance: (mame: `cas' a `uhel' prijezdu) otestujeme jestli jiz
 * nahodou neexistuje arival ktery by se stihl otocit jeste pred
 * `cas'em  na prislusny `uhel'. pokud neexistuje pridame tento arival
 * a je nutne pridat vakanci do `vac_bag'u
 *
 * technicke veci:
 *   zjistit prunik `path'u s usekem.
 *   pridani nove cesty do existujiciho sveta.
 *
 * konvence: intervaly jsou (beg, end> otevrene-uzavrene
 */

t_path_item::t_path_item(
		e_action ac,
		t_needle *from, t_needle *to,
		double tb, double te)
{
	action = ac;
	tend = te;
	tbeg = tb;
	xbeg = from->x;
	ybeg = from->y;
	xend = to->x;
	yend = to->y;
} /* t_path_item::t_path_item */

/*
 * r: time kdy ze protno
 */
double
t_path_item::colission(const t_path_item &p)
{
	if( /* test jestli se bounding-kvadry protinaji */
		tbeg > p.tend ||
		xbeg > p.xend ||
		ybeg > p.yend ||
		tend < p.tbeg ||
		xend < p.xbeg ||
		yend < p.ybeg )
	{
		return TIME_INFTY;
	}

	double Ax = p.xbeg, Ay = p.ybeg, At = p.tbeg; 
	double Bx = p.xend, By = p.yend, Bt = p.tend; 

	double Cx = xbeg, Cy = ybeg, Ct = tbeg; 
	double Dx = xend, Dy = yend, Dt = tend; 

	double t_ab =  Bt - At;
	double t_cd =  Dt - Ct;

	double Lx, Ly, Mx, My;
	double x, y, t;

	Lx =  Ax / t_ab - Bx / t_ab - Cx / t_cd + Dx / t_cd;
	Ly =  Ay / t_ab - By / t_ab - Cy / t_cd + Dy / t_cd;

	Mx =  Ax * At / t_ab - Bx * Bt / t_ab - Cx * Ct / t_cd + Dx * Dt / t_cd;
	My =  Ay * At / t_ab - By * Bt / t_ab - Cy * Ct / t_cd + Dy * Dt / t_cd;

	if( Lx == 0 && Ly == 0 ) 
	{
		return  My * My + Mx * Mx;
	}

	t = - ( Mx * Lx + My * Ly ) / ( Lx * Lx + Ly * Ly );

	if ( t > Bt || t > Dt || t < At || t < Ct )
	{/* mno a to je prave blbe: */
		return TIME_INFTY;
	}

	x = t * Lx + Mx ;
	y = t * Ly + My ;

	return x * x + y * y;

} /* colisson */

void
t_world::init_task(
	t_needle& abeg_needle,
	t_needle &aend_needle,
	double abeg_time,
	double abeg_angle)
{
	beg_needle = &abeg_needle;
	end_needle = &aend_needle;

	beg_time = abeg_time;
	beg_angle = abeg_angle;
} /* t_world::init_task */

void
t_world::solve()
{
	/* inicialize fst vac insert it in `vac_bag'*/

	vac_bag.push(&(beg_needle->vacs[0]));
	beg_needle->vacs[0].add_arival(beg_time,beg_angle);

	db("solve     :push(%i, %5.2lf)",
		vac_bag.top()->id,
		vac_bag.top()->beg_min);

	/* set last vac */
	last_vac_arive = TIME_INFTY;


	while( last_vac_arive >  vac_bag.top()->beg_min )
	{
		/* vyber vacation */
		solve_vac();
	}
} /* t_world::slove */

/*
 * vyresi vacation ktera je na topu `vac_bag'u
 */
void
t_world::solve_vac()
{
	t_vac *vac;
	t_needle *nee;

	t_adjs::iterator i;

	/* vyber nejdrivejsi vacation */
	assert(vac_bag.size() > 0 );
	vac = vac_bag.top();
	vac_bag.pop();

	db("solve_vac : pop(%i, %5.2lf)",
		vac->id,
		vac->beg_min);


	/* patreim teto needli */
	nee = vac->nee;

	/* projed vsechny sousedy a postupne je vyres */
	for(i = nee->adjs.begin(); i != nee->adjs.end() ; i++)
	{
		solve_vac_needle(*vac, **i);
	}
} /* t_world::solve_vac() */

/*
 * vyresi problem z vacance `vac' do nedle `dest'
 */
void
t_world::solve_vac_needle(t_vac& vac, t_needle& dest)
{
	//double dist;
	//double deltatime; 

	//double possible_beg, possible_end;
	t_vacs::iterator i;

	compute_position(*vac.nee, dest);

	/* spocitej src_to_dest_angle -- jak nejmensi cas je treba k natoceni na
	 * `dest' 
	 * dist vzalenost stavajici needle a `dest'
	 * delta  za jak dlouho se dojede.
	 */

	/* `src_to_dest_angle' `vac.end' jsou hranice */

	for(i = dest.vacs.begin(); i != dest.vacs.end(); i++)
	{
		if( possible_end >= (*i).beg && possible_beg <= (*i).end )
		{
			solve_vac_vac(vac, *i);
		}
	}
} /* t_world::solve_vac() */

/*
 * spocita pomocne promene vzhdlem k zdrojove needli a koncove needli
 */
void
t_world::compute_position(t_needle& src, t_needle& dest)
{
	double dx = dest.x - src.x;
	double dy = dest.y - src.y;

	dist = sqrt(dx * dx + dy * dy);
	delta_time = dist / velocity;

	assert(delta_time >= 0 );

	src_to_dest_angle = nee_to_nee_angle(src, dest);

	assert(src_to_dest_angle >= 0);
	
	/* `src_to_dest_angle' `posible_end' a `posible_beg'  */
	/* jeste spocitat `ari_time' a `ari_angle' */

} /* t_world::compute_position() */


/*
 * zkousime se z vacance `src' trefit do vacance `dest'
 * pomocne promene `dist' `delta_time' jiz jsou pripraveny
 */
void
t_world::solve_vac_vac(t_vac& src, t_vac& dest)
{
	double try_end;
	double i;


	cut_time = src.fst_time_in_angle(src_to_dest_angle);
	assert(cut_time >= src.beg);
	try_end = dest.end;

	if( cut_time > try_end ) {
		/* tyhle vakance se vubec nepotkaji */
		return;
	}

	if( is_free_path(cut_time, src, dest))
	{	/* trefa: da se projet hned po natoceni */
		update_vac(src, dest);
		return;
	}

	/* najdi nejdrivejsi volnou cestu */
	for(	i = cut_time + TIME_STEP_COARSE;
		i <= try_end ;
		i += TIME_STEP_COARSE )
	{
		if( is_free_path(i, src, dest))
		{
			/* zkusime zjemnit */
			solve_vac_vac_down(i, src, dest);
			return;
		}
	}
	/* FIXME: jeste je mozne vyzkouset primo `try_end' a ne jen
	 * nejblizsi nizsi
	 */
} /* t_world::solve_vac_vac */

/*
 * pokusi se zjemni nalezene minimum, `found_best' byl
 * nejlepsi cas nalezeny s krokem `TIME_STEP_COARSE'
 * nyni budeme postupovat s krokem `TIME_STEP_FINE'
 */
void
t_world::solve_vac_vac_down(double found_best, t_vac& src, t_vac& dest)
{
	double down;
	for(down = 0;(down += TIME_STEP_FINE) < TIME_STEP_COARSE;)
	{
		if( is_free_path(found_best - down, src, dest) )
		{
			update_vac(src, dest);
		}

	}
} /* t_world::solve_vac_vac_down() */

/*
 * uses `ari_angle'
 * spocitej pripadne zarad novy arival z src do dest
 */
void
t_world::update_vac(t_vac& src, t_vac& dest)
{
	int rt;
	double ari_time, ari_angle;

	/* spocitej ari_time a ari_vec */
	ari_angle = nee_to_nee_angle(*src.nee, *dest.nee);
	ari_time = delta_time + cut_time;


	rt = dest.add_arival(ari_time, ari_angle, &src);
	if( rt ) {
		/* je to vylepseni pridej */
		db("update_vec:push(%i, %5.2lf) ariT: %5.2lf ariA:%5.2lf",
			dest.id,
			dest.beg_min,
			ari_time,
			ari_angle);

		vac_bag.push(&dest);
		if( dest.nee  == end_needle ) {
			update_last_vac_arive(ari_time);
		}
	}
} /* t_world::update_vac() */

void
t_world::update_last_vac_arive(double ari_time)
{
	if( last_vac_arive > ari_time) {
		last_vac_arive = ari_time;
	}
} /* t_world::update_last_vac_arive */



/*
 *
 * prida arival do vakance, udrzi v konzistenci
 * beg_min
 * r: !0 iff je to vylepseni a ma se zaradit do `vac_bag'u
 *
 */

int
t_vac::add_arival(double ari_time, double ari_angle, t_vac *src)
{
	t_arivals::iterator i;
	db("trying ari[nee:%i] tim:%lf  ang:%lf", nee->id, ari_time, ari_angle);

	for(i = arivals.begin(); i != arivals.end();)
	{
		db("ari:tim:%lf ang:%lf", (*i).tim, (*i).ang);
		if( (*i).is_covered(ari_time, ari_angle) )
		{	/* remove `*i' */
			i = arivals.erase(i);
			db("`--> is covered ... continuing");
			continue;
		}
		if( (*i).covers(ari_time, ari_angle)  )
		{
			db("`--> covers :(");
			/* nic noveho pod sluncem takove reseni uz
			 * mame a lepsi
			 */
			return 0;
		}
		i++;
	}

	if( ari_time < beg_min ) {
		beg_min = ari_time;
	}
	db("`--> succes");

	arivals.push_back(t_arival(ari_time, ari_angle, src));

	return 1;
} /* t_vac::add_arival */

double
t_vac::fst_time_in_angle(double angle)
{
	double cur_time;
	double best_time = TIME_INFTY;

	for(t_arivals::iterator i=arivals.begin();
			i!=arivals.end();i++) {
		cur_time = t_world::rot_time( (*i).ang, angle) + (*i).tim;
		if( cur_time < best_time ) {
			best_time =  cur_time;
		}
	}
	return best_time;
} /* t_vac::fst_time_in_angle */


/*
 * !0 pokud v ceste z needle vacance `src' do `dest' neco stoji
 */
int
t_world::is_free_path(double begtime, t_vac& src, t_vac& dest)
{
	t_paths::iterator i;
	t_path::iterator  j;

	t_needle
		*nee_s = src.nee,
		*nee_d = dest.nee;


	t_path_item p(	nee_s->x, nee_s->y, begtime,
			nee_d->x, nee_d->y, begtime + delta_time);


	for(i = paths.begin(); i != paths.end(); i++)
	for(j = (*i).begin(); j != (*i).end(); j++)
	{
		if(  p.colission(*j) != TIME_INFTY )
		{
			return 0;
		}	
	}
	return 1;
} /* t_world::is_free_path() */

/*
 * kdy bude prislusny arival schopny byt namiren na angle
 */
double
t_arival::time_in_angle(double angle)
{
	return tim + t_world::rot_time(angle, ang);
} /* t_arival::time_in_angle */

/*
 * je arival pokryty prijezdem v `by_tim' z uhlu `by_ang'
 * (uhel je v <0,2*M_PI> )
 */
int
t_arival::is_covered(double by_tim, double by_ang)
{
	//return is_covered_lo(by_tim, by_ang, tim, ang);
	double rtime = by_tim + t_world::rot_time(by_ang, ang);
	return rtime < tim;
} /* t_arival::is_coverd */

/*
 * pokryva arival prijezd v `by_tim' z uhlu `by_ang'
 * (uhel je v <0,2*M_PI> )
 */
int
t_arival::covers(double cov_tim, double cov_ang)
{
	//return is_covered_lo(tim, ang,  by_tim, by_ang);
	double rtime = tim + t_world::rot_time(cov_ang, ang);
	//db("covers: rtime:%lf", rtime);
	return rtime <= cov_tim;
} /* t_arival::covers */

#if 0
/*
 * je arival (prijezd v `my_tim' uhel `my_ang') pokryty prijezdem
 * v `by_tim' z uhlu `by_ang' (uhel je v <0,2*M_PI> )
 */
int
t_arival::time_rotis_covered_lo(
	double by_tim,
	double by_ang,
	double my_tim,
	double my_ang)
{
	double time_in_my_angle;
	double delta_angle;


	delta_angle = my_ang - by_ang;
	if( delta_angle < 0 )  {
		/* upravit uhel do <0,2*M_PI> */
		delta_angle += 2.0 * M_PI;
	}
	if( delta_angle > M_PI ) {
		/* nebudu delat trikrat vlevovbok kdyz mohu udelat
		 * jednou vpravovbok 
		 */
		delta_angle = 2.0 * M_PI - delta_angle;
	}
	time_in_my_angle = delta_angle * t_world::rot_velocity + by_tim;
	return time_in_my_angle <= my_tim;

} /* t_arival::is_covered */
#endif

/*
 *  jaky je cas na otocku z `from_angle' do `to_angle'
 */
double
t_world::rot_time(double from_angle, double  to_angle)
{
	double delta_angle;
	delta_angle = from_angle - to_angle;
	if( delta_angle < 0 )  {
		/* upravit uhel do <0,2*M_PI> */
		delta_angle += 2.0 * M_PI;
	}
	if( delta_angle > M_PI ) {
		/* nebudu delat trikrat vlevovbok kdyz mohu udelat
		 * jednou vpravovbok 
		 */
		delta_angle = 2.0 * M_PI - delta_angle;
	}
	assert(delta_angle >= 0);
	assert(delta_angle <= 2.0 * M_PI);

	return delta_angle * rot_velocity;
} /* t_world::rot_time */

double
t_world::nee_to_nee_time(t_needle &fr, t_needle &to)
{
	double dx = fr.x - to.x;
	double dy = fr.y - to.y;

	return sqrt(dx * dx + dy * dy) / velocity;
} /* t_world::nee_to_nee_time */

double
t_world::nee_to_nee_angle(t_needle &fr, t_needle &to)
{
	double dx = to.x - fr.x;
	double dy = to.y - fr.y;
	double angle;

	if( dx < EPSILON  && dy < EPSILON) {
		return 0;
	}
	angle = atan2(dy, dx);

	return angle < 0 ? angle + 2.0 * M_PI : angle;
	
} /* t_world::nee_to_nee_angle */

/*
 *  najde prvni arival ktery pokryva  ari(`time' , `angle')
 */

t_arival*
t_needle::find_fst_ari_covers(double time, double angle)
{
	t_vacs::iterator i;

	for(i = vacs.begin(); i != vacs.end(); i++ )
	{
		if( ! (*i).arivals.empty() ) {
			/* FIXME: `empty' neni postacujici podminka  aby
			 * existoval pokryvajici arival 
			 */
			return (*i).find_fst_ari_covers(time, angle);
		}
	}
	return  NULL;
} /* t_needle::find_ari() */

/*
 *  najde prvni arival ktery pokryva  ari(`time' , `angle')
 */
t_arival*
t_vac::find_fst_ari_covers(double time, double angle)
{
	t_arivals::iterator i;

	for(i = arivals.begin(); i != arivals.end(); i++)
	{
		if(  (*i).covers(time, angle) )
		{
			return &(*i);
		}
	}
	/* algoritmus nenasel zadne reseni. proste reseni neexistuje,
 	 * smula, fakt, no co mam delat? neni. zadne maliny proste
 	 * naschledanou, je mi lito, opravdu. zeptejte se za tyden.
	 */
	throw( t_world::t_ex("no solution"));
} /* t_needle::find_ari() */


/*
 *  najde nejlepsi arival ktery pokryva  `angle')
 */
t_arival*
t_vac::find_fst_ari_covers_angle(double angle)
{
	t_arivals::iterator i;
	t_arival *ret = NULL;
	double best_time= INFTY;
	double cur_time;

	for(i = arivals.begin(); i != arivals.end(); i++)
	{
		cur_time = (*i).time_in_angle(angle) ;
		if(  cur_time < best_time )
		{
			best_time = cur_time;
			ret = &*i;
		}
	}

	if( ret == NULL )  {
		throw( t_world::t_ex("no solution"));
	}
	return ret;
} /* t_vac::find_fst_ari_covers_angle */

/*
 * z nalezene cesty je treba udelat cestu a postavit nove needle
 */
void
t_world::build_path()
{
	t_path *p = new t_path;

	t_needle *nee_cur, *nee_new;
	t_vac *vac_new;
	t_arival *ari_cur, *ari_new;

	double angle_cur, time_cur;
	double angle_new, time_new;
	
	double jump_time, rot_t;
	double time_dojezd, time_otoceni_end;
	
	/* v `dest_needle' je koncna stanice */
	/* v `dest_angle' je koncna stanice */
	time_cur = TIME_INFTY;
	angle_cur = 0.0;
	nee_cur = end_needle;

	//FIXME: zbytecne se prohledava posledni vacance
	//FIXME: navic to hleda blbe
	ari_cur = nee_cur->find_fst_ari_covers(
					time_cur,
					angle_cur);
	assert(ari_cur);

	while( nee_cur != beg_needle )
	{
		vac_new = ari_cur->from;

		//ari_new = vac_new->find_fst_ari_covers(
					//time_cur,
					//angle_cur);

		nee_new = vac_new->nee;

		jump_time = nee_to_nee_time(*nee_new, *nee_cur);
		angle_new = nee_to_nee_angle(*nee_new, *nee_cur);
		rot_t =  rot_time(angle_cur, angle_new); 

		//time_new = time_cur - rot_t - jump_time;

		ari_new = vac_new->find_fst_ari_covers_angle(
					angle_new);
		time_new = ari_new->time_in_angle(angle_new);

		time_dojezd = time_new + jump_time;
		
		time_otoceni_end = time_dojezd + rot_t;

		/* cekani na volnou cestu */
		p->push_front(
			t_path_item(
				t_path_item::wait,
				nee_cur, nee_cur,
				time_otoceni_end , time_cur));
		/* nataceni se na dalsi */
		p->push_front(
			t_path_item(
				t_path_item::rot,
				nee_cur, nee_cur,
				time_dojezd, time_otoceni_end));
		/*cesta z new do cur */
		p->push_front(
			t_path_item(
				t_path_item::walk,
				nee_new, nee_cur,
				time_new, time_dojezd));
		time_cur = time_new;
		angle_cur = angle_new;
		nee_cur = nee_new;
		ari_cur = ari_new;
	}

	rot_t = rot_time(angle_cur, beg_angle); 
	time_otoceni_end = beg_time + rot_t; 

	/* cekani na volnou cestu */
	p->push_front(
		t_path_item(
			t_path_item::wait,
			nee_cur, nee_cur,
			time_otoceni_end , time_cur));

	/* nataceni se na dalsi */
	p->push_front(
		t_path_item(
			t_path_item::rot,
			nee_cur, nee_cur,
			beg_time, time_otoceni_end));

	paths.push_back(*p);
} /* t_world::buildpath() */

/*
 * vycisti vsechny vakance od vsech arivalu
 */
void
t_world::clean_vac()
{
	t_needles::iterator i;
	for(i = needles.begin(); i != needles.end(); i++) {
		(*i)->clean_vac();
	}
} /* t_world::clean_vac */

void
t_needle::clean_vac()
{
	t_vacs::iterator i;
	for(i = vacs.begin(); i != vacs.end(); i++) {
		(*i).clean_vac();
	}

} /* t_neelde::clean_vac */

void
t_vac::clean_vac() 
{
	arivals.erase(arivals.begin(), arivals.end());
} /* t_vac::clean_vac */

void
t_path_item::print()
{
	
	switch(action) {
	case wait:
		printf("pitem: wait[%5.2lf, %5.2lf]", tbeg, tend);
		printf(" @ (%5.2lf, %5.2lf) w\n", xbeg, ybeg);
		break;
	case noth:
		printf("pitem: noth\n");
		break;
	case rot:
		printf("pitem: rott[%5.2lf, %5.2lf]", tbeg, tend);
		printf(" @ (%5.2lf, %5.2lf) r\n", xbeg, ybeg);
		break;
	case walk:
		printf("pitem: walk[%5.2lf, %5.2lf]", tbeg, tend);
		printf(" @ (%5.2lf, %5.2lf) -> ", xbeg, ybeg);
		printf("(%5.2lf, %5.2lf) wl\n", xend, yend);
		break;
	default:
		printf("pitem: err: call vitas\n", tbeg, tend);

	}
} /* t_path_item::print */
/*
 * vytiskne se posledni cesta
 */

void
t_world::print_path()
{
	t_path &p = paths[0];
	t_path::iterator i;

	for(i=p.begin(); i != p.end(); i++) {
		(*i).print();
		
	}

	
} /* t_world::print_path */

void
t_world::split_vacs_by_last_path(){
	split_vacs_by_path(paths.back());
}

void
t_world::split_vacs_by_path(t_path &p)
{
	t_needles::iterator i;

	for(i = needles.begin(); i != needles.end(); i++) {
		(*i)->split_vacs_by_path(p);
	}
} /* t_world::split_vacs_by_path */


void
t_needle::split_vacs_by_path(t_path &p)
{
	t_path::iterator pi;

	for(pi = p.begin(); pi != p.end(); pi++) {
		split_vacs_by_path_item(*pi);
	}
} /* t_needle::split_vacs_by_path */

/*
 * zkonzistentni vac v needle pridanim cesty
 */
void
t_needle::split_vacs_by_path_item(t_path_item& p)
{
	t_time_pair tp;
	t_vacs::iterator i;

	for(i = vacs.begin(); i != vacs.end(); ) {

		tp = (*i).find_intersection(p);

		if( tp.end < (*i).beg || tp.beg > (*i).end ) {
			/* cele mimo */
			i++;
			continue;
		}

		if( tp.beg <= (*i).beg && tp.end >= (*i).end ) {
			/* cele to zakryva */
			i = vacs.erase(i);
			continue;
		}

		if( tp.beg > (*i).beg && tp.end < (*i).end ) {
			/* je treba sekat na dva */
			i = split_vac_by_time_pair(i, tp);
		} else {
			/* jen je treba posoupnout hranice */

			i = corect_vac_borders(i, tp);
		}
	}

} /* update_vac() */


/*
 *         v- beg   v- end
 * *i:     #########
 * tp:  xxxxxx 
 * resp:        xxxxxxx
 *
 *            %%%%%%
 * resp:   %%%%%
 *
 */
t_vacs::iterator
t_needle::corect_vac_borders(t_vacs::iterator i, t_time_pair tp)
{
	if( (*i).beg < tp.end ) {
		(*i).beg = tp.end;
	} else if( (*i).end > tp.beg ) {
		(*i).end = tp.beg;
	} else {
		assert(0);
	}


	return i + 1;
} /* t_needle::corect_vac_borders */

/* 
 *     v- beg        v- end
 * *i: ##############
 * tp:      xxxxx
 *
 * todo:
 * *i: %%%%%
 *     ^b1  ^e1
 * *(i+1):       %%%%
 *               ^b2 ^e2
 *
 * r: next i
 *
 */

t_vacs::iterator
t_needle::split_vac_by_time_pair(t_vacs::iterator i, t_time_pair tp)
{
	double b1, e1, b2, e2;

	b1 = (*i).beg;
	e1 = tp.beg;

	b2 = tp.end;
	e2 = (*i).end;

	assert( b1 <= e1);
	assert( e1 <= b2);
	assert( b2 <= e2);

 	(*i).beg = b2;
 	(*i).end = e2;

	i = vacs.insert(i, t_vac(this, b1, e1));

	return i + 2;
} /* t_needle::split_vac_by_time_pair */

/*
 * najde kdy se protne casoprostorocara `p'
 * (s kruhovym okolim o polomeru t_world::disk_sq )
 * s bodem nee->x,y
 */



t_time_pair
t_vac::find_intersection(t_path_item& p)
{
	double disk_sq = t_world::disk_sq;


	double
		pc, /* vzalenost stredu pruniku */
		pd,   /* vzalenost kraju pruniku od stredu */
		pb, pe, /* prvni a druhy prunik */
		Rx, Ry, /* A -> stred pruniku */
		Rsq, /* velikost ^2 */
		Vabs, /* velikost V (z B do E) */
		Vsq,
		Vnorm_x, /* normalizovane V */
		Vnorm_y;
	double
		Tbeg, Tend, /* vysledny interval */
		tb = p.tbeg,
		te = p.tend;


	double
		/* stred kruznice */
		Ax = nee->x,
		Ay = nee->y,

		/* pocatecni bod */
		Bx = p.xbeg,
		By = p.ybeg,

		/* koncovy */
		Ex = p.xend,
		Ey = p.yend,

		/* vektor V = E - B */
		Vx = Ex - Bx,
		Vy = Ey - By,

		/* z pocatku do A, U = A - B */
		Ux = Ax - Bx,
		Uy = Ay - By,
		Usq;


	Vsq  = Vx * Vx + Vy * Vy;
	
//cout << "intersec:[" << Bx << "," << By << "]@" << tb << "->"; 
//cout <<          "[" << Ex << "," << Ey << "]@" << te << " x "; 
//cout <<          "[" << Ax << "," << Ay << "] = "; 

	/* fixme suld be epsilon */
	if( Vsq <  0.00001)  {
		/* robot stoji na miste: zjisti jak daleko  */
		Usq = Ux * Ux + Uy * Uy;
		if( Usq > disk_sq ) {
			/* je daleko: zadny prunik */
//cout << "--\n";
			return t_time_pair(-INFTY, -INFTY);
		
		} 

		/*  je blizko: vrat presne dobu pdle `p' */
//cout << "[" << tb << "," << tb << "]\n"; 
		
		return t_time_pair(p.tbeg, p.tend);
	}

	Vabs = sqrt(Vsq);

	Vnorm_x = Vx / Vabs;
	Vnorm_y = Vy / Vabs;

	pc = Vnorm_x * Ux + Vnorm_y * Uy;

	Rx = Vnorm_x * pc - Ux;
	Ry = Vnorm_y * pc - Uy;

	Rsq = Rx * Rx + Ry * Ry;

	if( Rsq > disk_sq ) {
		/* je daleko: zadny prunik */
//cout << "--\n";
		return t_time_pair(-INFTY, -INFTY);
	} 

	pd = sqrt( disk_sq - Rsq );

	pb = pc - pd;
	pe = pc + pd;

	Tbeg = ( pb / Vabs ) * (te - tb) + tb;
	Tend = ( pe / Vabs ) * (te - tb) + tb;

//cout << "[" << Tbeg << "," << Tend << "]\n"; 

	return t_time_pair(Tbeg, Tend, tb, te);

} /* t_vac::find_intersection() */

double t_world::velocity = 1.0 ;
double t_world::rot_velocity = 1.0 ;

int 
main()
{
	try {
	t_needle n0(0,0);
	t_needle n1(0,5);
	t_needle n2( 5,10);
	t_needle n3(-5,10);
	t_needle n4( 0,15);
	t_needle n5(-5,15);

	n0.adjs.push_back(&n1);

	n1.adjs.push_back(&n0);
	n1.adjs.push_back(&n2);
	n1.adjs.push_back(&n3);

	n2.adjs.push_back(&n1);
	n2.adjs.push_back(&n4);

	n3.adjs.push_back(&n1);
	n3.adjs.push_back(&n4);

	n4.adjs.push_back(&n2);
	n4.adjs.push_back(&n3);
	n4.adjs.push_back(&n5);

	n5.adjs.push_back(&n4);

	t_world w;

	w.needles.push_back(&n0);
	w.needles.push_back(&n1);
	w.needles.push_back(&n2);
	w.needles.push_back(&n3);
	w.needles.push_back(&n4);
	w.needles.push_back(&n5);

	w.init_task(n0, n5, 0);
	w.solve();
	w.build_path();
	w.clean_vac();
	w.split_vacs_by_last_path();
	w.print_needles();
	w.print_path();


	} catch(t_world::t_ex e) {
		cout << "catch:" << e.msg;
	}

	return 0;
} /* main */
