#include <vector>
#include <string>
#include <queue>
#include <slist>
#include <iterator>
#include <math.h>
#include "verr.h"

#define INFTY 1000
#define EPSILON .000001
#define TIME_INFTY 10000
#define TIME_STEP_COARSE .1
#define TIME_STEP_FINE .02

class t_needle;
class t_vac;


class t_arival{
public:
	double tim;
	double ang; /* uhel bude udrzovan v rozmezi <0, 2*M_PI> */
	t_vac *from;
	t_arival(double atim, double aang, t_vac *afrom)
		:tim(atim), ang(aang), from(afrom){
		assert(ang >= 0);
		assert(ang <= M_PI * 2.0);
		if(!from){db("t_arival:!from");}
	};

	int is_covered(double time, double angle);
	int covers(double time, double angle);
	void print() { 
		printf("    ari: tim:%.2lf ang:%.2lf\n", tim, ang);
	};
	double time_in_angle(double angle);
	
private:
	//int is_covered_lo(double time, double angle,
	//	double my_tim, double my_ang);
}; /* class t_arival */

class t_path_item{
public:
	enum e_action {
		noth,
		walk,
		rot,
		wait,
	};

	t_path_item(double xb, double yb, double tb, 
		double xe, double ye, double te)
	:action(noth), xbeg(xb), ybeg(yb), tbeg(tb), xend(xe), yend(ye), tend(te)
	{};

	t_path_item(e_action ac, t_needle *from, t_needle *to,
		double tb, double te);
	double colission(const t_path_item &p);
	void print();

	double xbeg, ybeg, tbeg;
	double xend, yend, tend;
	e_action action;
	
		
};/* class t_path_item */

/*
 * konvence: intervaly jsou (beg, end> otevrene-uzavrene
 */

class t_time_pair {
public:
	t_time_pair(double b = 0.0, double e = 0.0):beg(b),end(e){};

	/* interval <b,e) se orizen <orez_b, orez_e) */
	t_time_pair(double b, double e, double orez_b, double orez_e){

		if( e < orez_b || b > orez_e ) {
			beg = -INFTY;
			end = -INFTY;
		} else {
			beg = b < orez_b ? orez_b : b;
			end = e > orez_e ? orez_e : e;
		}
	}
	double beg, end;
};
class t_path: public vector<t_path_item>{
public:
	void push_front(const t_path_item &p) { insert(begin(), p); };
};


typedef vector<t_path> t_paths;

typedef vector<t_arival> t_arivals;

class t_vac{
	public:
	int id;
	double beg, end; /* od kdy do kdy je tato needle volna */
	//double beg_full; /* kdy zacina plne pokryti kruhu */ 
	double beg_min; /* kdy prijede vubec prvni */
	t_needle *nee;
	t_arivals arivals;
	t_vac(t_needle *my_nee, double b=0, double e=TIME_INFTY)
		:beg(b),end(e),nee(my_nee),
		//beg_full(TIME_INFTY),
		beg_min(TIME_INFTY),
		id(id_top++){
			assert((int)my_nee > 0x80000000 );
		};

	int operator< (t_vac &vac){ db("op<");return beg_min > vac.beg_min; };

	/* najde nejlepsi arival pro tento azimut */
	t_arival* find_fst_ari_covers(double time, double angle);
	t_arival* find_fst_ari_covers_angle(double angle);
	t_time_pair  find_intersection(t_path_item& p);
	void update_vac(t_path& p);
	void print() {
		printf("  vac beg:%5.2lf end:%5.2lf [\n", beg, end);
		for(t_arivals::iterator i=arivals.begin();
			i!=arivals.end();
			i++) {
			(*i).print();
		}
		printf("  ]\n");
	};
	int add_arival(
			double ari_time,
			double ari_angle,
			t_vac *src = NULL);
	double fst_time_in_angle(double angle);

	void clean_vac();
private:
	static int id_top;
		

};/* class t_vac */

class t_vac_cmp {
public:
	bool operator()(t_vac *a, t_vac*b) {
		return a->beg_min > b->beg_min;
	};
}; /* class t_vac_cmp */

#if 0
class t_vacp {
public: 
	t_vac *p;
	t_vacp(t_vac *pp = NULL):p(pp){};
	int operator<(t_vacp &pp) {
		return *p < *(pp.p);
	};
	t_vac & operator *() {
		return *p;
	};
};
#endif
//operator < (t_vac *v1, t_vac*v1);

typedef vector<t_needle*> t_adjs;
typedef vector<t_vac> t_vacs;


class t_needle {
	public:
	double x, y;
	t_adjs adjs;
	t_vacs vacs;
	t_needle(double xx=0, double yy=0):x(xx),y(yy), id(id_top++){
		vacs.push_back(t_vac(this));
	};

	/* najde minimalni vac ktera obsahuje nejaky arival */
	t_arival* find_fst_ari_covers(double time, double angle);
	void print(){
		printf("%i(%5.2lf,%5.2lf)\n", id, x, y);
		for(t_vacs::iterator i=vacs.begin();i!=vacs.end();i++) {
			(*i).print();
		}
		printf("\n\n");
	};
	void clean_vac();
	void split_vacs_by_path(t_path &p);
	void split_vacs_by_path_item(t_path_item &pi);

	t_vacs::iterator corect_vac_borders(t_vacs::iterator i, t_time_pair tp);
	t_vacs::iterator split_vac_by_time_pair(
				t_vacs::iterator i,
				t_time_pair tp);

	int id;
private:
	static int id_top;

}; /* class t_needle */

typedef vector<t_needle*> t_needles;



class t_world {
public:
	class t_ex {
		public:
		string msg;
		t_ex(char *s){msg = s;};
	};
	static double velocity; /* rychlost robota */
	static double rot_velocity; /* ryclost otaceni */
	t_needle *beg_needle;
	t_needle *end_needle;
	double beg_time;
	double beg_angle;

	/* pom pri pri reseni nee_to_nee, predpocitany azimut*/
	double src_to_dest_angle;

	/* pom vychozi cas z vacance, nastavi `solve_vac_vac()' */
	double cut_time; 
	double possible_beg, possible_end;
	/* FIXME: co to znamena delta_time? */
	double delta_time;
	/* FIXME: co to znamena dist? */
	double dist;

	/* disk -- polomer robota */
	static double disk;
	static double disk_sq;

	double ari_time;
	double ari_angle;
	t_paths paths;
	t_needles needles;
public:
	typedef
		priority_queue<t_vac*,vector<t_vac*>, t_vac_cmp>
		t_vacbag;

	t_vacbag vac_bag;
	double last_vac_arive;

	static double rot_time(double a1, double a2);
	static double nee_to_nee_time(t_needle &fr, t_needle &to);
	static double nee_to_nee_angle(t_needle &fr, t_needle &to);

	void solve();
	void solve_vac();
	void solve_vac_needle(t_vac& vac, t_needle& dest);
	void solve_vac_vac(t_vac& src, t_vac& dest);
	void solve_vac_vac_down(
			double f,
			t_vac& src,
			t_vac& dest);
	void update_vac(t_vac& src, t_vac& dest);
	void update_last_vac_arive(double ari_time);
	int  is_free_path(
			double beg,
			t_vac& src,
			t_vac&dest);
	void compute_position(t_needle& src, t_needle& dest);
	void build_path(); 
	void print_path();
	void print_needles() {
		for(t_needles::iterator i=needles.begin();
			i!=needles.end();i++){
				(*i)->print();
		}
	};
	void t_world::init_task(
		t_needle& abeg_needle,
		t_needle &aend_needle,
		double abeg_time = 0 ,
		double abeg_angle = 0);
	void split_vacs_by_path(t_path &p);
	void split_vacs_by_last_path();
	void clean_vac();
}; /* class t_world */

